/***********************************************************************************************************************
 *
 * SolidBlue - open source safe data
 * Copyright (C) 2011-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://solidblue.java.net
 * SCM: https://bitbucket.org/tidalwave/solidblue-src
 *
 **********************************************************************************************************************/
package it.tidalwave.actor.spi;

import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import javax.inject.Provider;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import it.tidalwave.messagebus.MessageBus.Listener;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.actor.annotation.Actor;
import it.tidalwave.actor.annotation.CollaborationListener;
import it.tidalwave.actor.annotation.Message;
import it.tidalwave.actor.annotation.MessageListener;
import it.tidalwave.actor.Collaboration;
import it.tidalwave.actor.CollaborationCompletedMessage;
import it.tidalwave.actor.CollaborationStartedMessage;
import it.tidalwave.actor.impl.DefaultCollaboration;
import it.tidalwave.actor.impl.ExecutorWithPriority;
import java.lang.reflect.InvocationTargetException;
import java.util.Collections;
import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import lombok.Getter;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.openide.util.Exceptions;

/***********************************************************************************************************************
 * 
 * @stereotype Actor
 * 
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
@Slf4j @RequiredArgsConstructor
public class ActorActivator
  {
    @Getter
    private final Object actorObject;
    
    private final Provider<CollaborationAwareMessageBus> messageBus = Locator.createProviderFor(CollaborationAwareMessageBus.class);
    
    private final List<Listener<?>> messageBusListeners = new ArrayList<Listener<?>>();
    
    private final ExecutorWithPriority executor;

    /*******************************************************************************************************************
     * 
     * 
     * 
     ******************************************************************************************************************/
    @RequiredArgsConstructor
    class MessageListenerAdapter<Topic> implements Listener<Topic>
      {
        @Nonnull
        private final Method method;
        
        @Override
        public void notify (final @Nonnull Topic message) 
          {
            log.trace("notify({})", message);
            final DefaultCollaboration collaboration = DefaultCollaboration.getCollaboration(message);
            collaboration.registerPendingMessage(message); 
            
            final Runnable messageWorker = new Runnable()
              {
                @Override
                public void run() 
                  {
                    collaboration.bindToCurrentThread();
                    collaboration.unregisterPendingMessage(message);

                    try
                      {
                        method.invoke(actorObject, message);
                      }
                    catch (Throwable t)
                      {
                        log.error("Error calling {} with {}", method, message.getClass());
                        log.error("", t); 
                      }
                    finally
                      {
                        collaboration.unbindFromCurrentThread();
                      }
                  }
              };

            if (message.getClass().getAnnotation(Message.class).outOfBand())
              {
                executor.executeWithPriority(messageWorker);
              }
            else
              {
                executor.execute(messageWorker);
              } 
          }
      }
    
    /*******************************************************************************************************************
     * 
     * 
     * 
     ******************************************************************************************************************/
    @RequiredArgsConstructor
    class CollaborationMessageListenerAdapter<Topic extends Collaboration.Provider> implements Listener<Topic>
      {
        @Nonnull
        private final Method method;
        
        @Nonnull
        private final Class<?> type;
        
        @Override
        public void notify (final @Nonnull Topic message) 
          {
            log.trace("notify({})", message);
            final DefaultCollaboration collaboration = (DefaultCollaboration)message.getCollaboration();
            collaboration.registerPendingMessage(message); 
            
            final Runnable messageWorker = new Runnable()
              {
                @Override
                public void run() 
                  {
                    if (!collaboration.getOriginatingMessage().getClass().equals(type))
                      {
                        collaboration.unregisterPendingMessage(message);                        
                      }
                    else
                      {
                        collaboration.bindToCurrentThread();
                        collaboration.unregisterPendingMessage(message);

                        try
                          {
                            method.invoke(actorObject, message);
                          }
                        catch (Throwable t)
                          {
                            log.error("Error calling {} with {}", method, message.getClass());
                            log.error("", t); 
                          }
                        finally
                          {
                            collaboration.unbindFromCurrentThread();
                          }
                      }
                  }
              };

            executor.execute(messageWorker);
          }
      }
    
    /*******************************************************************************************************************
     * 
     * 
     * 
     ******************************************************************************************************************/
    public ActorActivator (final @Nonnull Class<?> actorClass, final @Nonnegative int poolSize)
      {
        try
          {
            final Actor actor = actorClass.getAnnotation(Actor.class);

            if (actor == null)
              {
                throw new IllegalArgumentException("Actor class must be annotated with @Actor: " + actorClass);  
              }

            if (!actor.threadSafe() && (poolSize != 1))
              {
                throw new IllegalArgumentException("Actors that aren't thread safe can't have pool size > 1");  
              }

            actorObject = actorClass.newInstance();
            executor = new ExecutorWithPriority(poolSize, actorClass.getSimpleName(), actor.initialPriority());
          }
        catch (InstantiationException e)
          { 
            throw new RuntimeException(e);  
          }
        catch (IllegalAccessException e)
          {
            throw new RuntimeException(e);  
          }
      }
    
    /*******************************************************************************************************************
     * 
     * 
     * 
     ******************************************************************************************************************/
    public void initialize() 
      {        
        for (final Class<?> clazz : getClassHierarchy(actorObject.getClass()))
          {
            for (final Method method : clazz.getDeclaredMethods())
              {
                if (method.getAnnotation(MessageListener.class) != null)
                  {
                    registerMessageListener(method);
                  }
                else if (method.getAnnotation(CollaborationListener.class) != null)
                  {
                    registerCollaborationListener(method);
                  }
                else if (method.getAnnotation(PostConstruct.class) != null)
                  {
                    try 
                      {
                        method.invoke(actorObject);
                      }
                    catch (IllegalAccessException e) // TODO: use a checked exception
                      {
                        throw new RuntimeException(e);
                      }
                    catch (IllegalArgumentException e) 
                      {
                        throw new RuntimeException(e);
                      }
                    catch (InvocationTargetException e) 
                      {
                        throw new RuntimeException(e);
                      }
                  }
              }
          }
      }

    /*******************************************************************************************************************
     * 
     * 
     * 
     ******************************************************************************************************************/
    public void dispose() 
      {
        for (final Listener<?> listener : messageBusListeners)
          {
            messageBus.get().unsubscribe(listener);  
          }
        
        final List<Class> hierarchy = getClassHierarchy(actorObject.getClass());
        Collections.reverse(hierarchy);
        
        for (final Class<?> clazz : hierarchy)
          {
            for (final Method method : clazz.getDeclaredMethods())
              {
                if (method.getAnnotation(PreDestroy.class) != null)
                  {
                    try 
                      {
                        method.invoke(actorObject);
                      }
                    catch (IllegalAccessException e) // TODO: use a checked exception
                      {
                        throw new RuntimeException(e);
                      }
                    catch (IllegalArgumentException e) 
                      {
                        throw new RuntimeException(e);
                      }
                    catch (InvocationTargetException e) 
                      {
                        throw new RuntimeException(e);
                      }
                  }
              }
          }
      }

    /*******************************************************************************************************************
     * 
     * 
     * 
     ******************************************************************************************************************/
    private <Topic> void registerMessageListener (final @Nonnull Method method)
      {
        log.info("registerMessageListener({})", method);
        
        if ((method.getParameterTypes().length != 1)
            || (method.getParameterTypes()[0].getAnnotation(Message.class) == null))
          {
            throw new IllegalArgumentException("Methods annotated as @MessageListener must have exactly "
                                             + "one parameter whose type is annotated with @Message: " + method);  
          }

        addListener(method, new MessageListenerAdapter<Topic>(method), (Class<Topic>)method.getParameterTypes()[0]);        
      }

    /*******************************************************************************************************************
     * 
     * 
     * 
     ******************************************************************************************************************/
    private void registerCollaborationListener (final @Nonnull Method method)
      {
        log.info("registerCollaborationListener({})", method);
        final Class<?> parameterType = method.getParameterTypes()[0];
        
        if ((method.getParameterTypes().length != 1)
            || (!parameterType.equals(CollaborationStartedMessage.class) && !parameterType.equals(CollaborationCompletedMessage.class)))
          {
            throw new IllegalArgumentException("Methods annotated as @CollaborationListener must have exactly "
                                             + "one parameter whose type is CollaborationStartedMessage or CollaborationCompletedMessage: " + method);  
          }

        final Class<?> type = method.getAnnotation(CollaborationListener.class).message();
        final Listener messageListener = parameterType.equals(CollaborationStartedMessage.class)
                                               ? new CollaborationMessageListenerAdapter<CollaborationStartedMessage>(method, type)
                                               : new CollaborationMessageListenerAdapter<CollaborationCompletedMessage>(method, type);
        addListener(method, messageListener, parameterType);        
      }

    /*******************************************************************************************************************
     * 
     * 
     * 
     ******************************************************************************************************************/
    private <Topic> void addListener (final @Nonnull Method method, 
                                      final @Nonnull Listener<Topic> messageListener,
                                      final @Nonnull Class<Topic> topic) throws SecurityException 
      {
        method.setAccessible(true);
        messageBusListeners.add(messageListener);
        messageBus.get().subscribe(topic, messageListener);
      }  
    
    /*******************************************************************************************************************
     * 
     * Returns the hierarchy of the given class, starting from the highest.
     * 
     ******************************************************************************************************************/
    @Nonnull
    private static List<Class> getClassHierarchy (final @Nonnull Class<?> clazz)
      {  
        final List<Class> hierarchy = new ArrayList<Class>();
        
        for (Class<?> ancestor = clazz; ancestor != null; ancestor = ancestor.getSuperclass())
          {
            hierarchy.add(0, ancestor);
          }
        
        return hierarchy;
      }
  } 
